# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import defaultdict
import sys

from colcon_core.event.job import JobEnded
from colcon_core.event.output import StderrLine
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version


class ConsoleStderrEventHandler(EventHandlerExtensionPoint):
    """
    Output all stderr of a task at once.

    The output is batched up until the task has ended in order to not
    interleave the output from parallel tasks.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.output.StderrLine`
    - :py:class:`colcon_core.event.job.JobEnded`
    """

    # the priority should be slightly higher than the default priority
    # in order to output errors early
    PRIORITY = 110

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._stderr_lines = defaultdict(list)

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, StderrLine):
            job = event[1]
            self._stderr_lines[job].append(data.line)

        elif isinstance(data, JobEnded):
            job = event[1]
            if self._stderr_lines[job]:
                msg = '--- stderr: {data.identifier}\n' \
                    .format_map(locals()) + \
                    b''.join(
                        self._stderr_lines[job]).decode() + \
                    '---'
                print(msg, file=sys.stderr, flush=True)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import time

from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.location import create_log_path
from colcon_core.location import get_log_path
from colcon_core.plugin_system import satisfies_version


class EventLogEventHandler(EventHandlerExtensionPoint):
    """
    Log all events to a global log file.

    The log file `events.log` is created in the log directory.
    """

    FILENAME = 'events.log'

    # the priority should be higher than the default priority
    # in order to write the events to the log file first
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._path = None
        self._start_time = None

    def __call__(self, event):  # noqa: D102
        data = event[0]

        self._init_log()

        context = str(event[1]) if event[1] is not None else '-'
        try:
            members = data.__dict__
        except AttributeError:
            members = {s: getattr(data, s) for s in data.__slots__}
        with self._path.open(mode='a') as h:
            h.write(
                '[%f] (%s) %s: %s\n' % (
                    self._get_relative_time(), context,
                    data.__class__.__name__, members))

    def _init_log(self):
        # only create log once
        if self._path is not None:
            return

        create_log_path(self.context.args.verb_name)
        self._path = get_log_path() / EventLogEventHandler.FILENAME
        with self._path.open(mode='w'):
            pass

    def _get_relative_time(self):
        now = time.monotonic()
        if self._start_time is None:
            self._start_time = now
            return 0
        return now - self._start_time



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import locale

from colcon_core.event.job import JobEnded
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_output.event_handler.log import get_log_directory
from colcon_output.event_handler.log import STDOUT_STDERR_LOG_FILENAME


class ConsoleCohesionEventHandler(EventHandlerExtensionPoint):
    """
    Pass task output at once to stdout.

    The extension (indirectly) handles events of the following types:
    - :py:class:`colcon_core.event.command.Command`
    - :py:class:`colcon_core.event.output.StdoutLine`
    - :py:class:`colcon_core.event.output.StderrLine`
    """

    # this handler is disabled by default
    # in favor of he `console_direct` or `console_log` handler
    # but other handlers might choose to change that presetting
    ENABLED_BY_DEFAULT = False

    # the priority should be lower than the `log` handler
    # since it reads the log file written by that extension
    # the priority should be higher than the `console_stderr` handler
    # in order for the `stderr` output to appear below the combined output
    PRIORITY = 130

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self.enabled = ConsoleCohesionEventHandler.ENABLED_BY_DEFAULT

    def __call__(self, event):  # noqa: D102
        data = event[0]

        # instead of buffering all output in memory
        # read the combined log file written by the `log` event handler
        if isinstance(data, JobEnded):
            job = event[1]
            base_path = get_log_directory(job)
            if not (base_path / STDOUT_STDERR_LOG_FILENAME).exists():
                return

            # ensure that output is printable on the console
            encoding = locale.getpreferredencoding()
            with (base_path / STDOUT_STDERR_LOG_FILENAME).open(
                mode='r', encoding=encoding, errors='replace'
            ) as h:
                content = h.read()

            msg = '--- output: {data.identifier}\n'.format_map(locals()) + \
                content + \
                '---'
            print(msg, flush=True)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.event.job import JobQueued
from colcon_core.event.job import JobStarted
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version


class ConsolePackageListEventHandler(EventHandlerExtensionPoint):
    """
    Output list of queued job names.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.job.JobQueued`
    - :py:class:`colcon_core.event.job.JobStarted`
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self.enabled = False
        self._queued = []

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, JobQueued):
            job = event[1]
            self._queued.append(job)
        if isinstance(data, JobStarted):
            if self._queued:
                print('Topological order')
                for job in self._queued:
                    print(
                        '- {job} ({job.task_context.pkg.type})'
                        .format_map(locals()))
                # clear the list to only print the order once
                self._queued = []
